/******************************************************************************
*                                                                             *
*                  Outil de comparaison des hashs de fichiers                 *
*                                                                             *
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>

#include "cmphash.h"
#include "delh.h"

char *prog_name;
Options opt;

void version()
{
  fprintf(stdout,"%s, %s - %s - Version %s\n",prog_name,Auteur,Date,Version);  
}

int usage(int status)
{
 if (status != 0)
    {
      fprintf (stderr, 
	       "Pour en savoir davantage, faites: `%s --help'\n",prog_name);
    }
  else
    {
      version();
      fprintf(stdout,"\
Permet d'effectuer des comparaisons entre diffrentes collections de signatures.\n\
Il s'agit de comparer des hashs de fichiers (md5, sha1, ...) connus avec \n\
d'autres inconnus (FICHIER).\n");
      fprintf(stdout,"\nUsage : \n\
  %s [options] FICHIER\n",prog_name);
      fprintf(stdout,"\nOptions :\n\
  -h, --help\t\tCette aide\n\
  -v, --version\t\tAffiche la version du programme\n\
  -s, --set=REPERTOIRE\tCharge les Hash sets du rpertoire en question\n\
  -0, --print0\t\tL'affichage des fichiers se termine par le caractre 0\n\
  -c, --col=nombre\tIndique les colonnes  afficher : col1=4, col2=2, col3=1\n\
  -m, --match\t\tAffiche uniquement les fichiers connus (par dfaut)\n\
  -r, --reject\t\tAffiche uniquement les fichiers inconnus\n\
  -t, --stat\t\tAffiche une ligne de statistique\n\
  ");
      fprintf(stdout,"\nSortie :\n\
  En fonction des options choisies.\n");
    }
  return TRUE;
}


ligne *lit_ligne(int fileno, ligne *l)
{
  /* Lecture d'une ligne du fichier fileno, et transformation de cette 
     ligne pour la faire entrer dans le moule de la structure de donne.
     on utilise cette mthode barbare car il existe des caractres spciaux
     dans les noms de fichiers, comme par exemple " ", "*", ... et fscanf ne
     ragit pas trs bien  ce type de caractres (surtout les espaces et les
     tabulations ...) */
  int i=0;
  int j=0;
  int k=0;
  char c[2] = "\0\0";

 
  while ((l->ok == TRUE) && (c[0] != '\n') && (i<(MAX_FILENAME+MAX_HASH_SIZE+16)))
    {
      if (read(fileno,c,1)==1)
	{
	  l->line[i] = c[0];
	  i++;
	}
      else
	l->ok=FALSE;
    }
  /* i est  la fin de la ligne (aprs le \n) */

  if (i==(MAX_FILENAME+MAX_HASH_SIZE+16)) /* pas trop loin non plus ! */
    l->ok=FALSE;
    
  if (i>6 && l->ok==TRUE) /* on doit lire au moins "MD5()= " */
    {
      l->line[i-1] = (char) 0; /* on remplace le \n par \0  */

      while ((i>0) && (l->line[i]!=' ')) /* on recherche le dbut du hash */
	i--;
     
      /* on extrait le hash (i est au premier chiffre hexadcimal du hash) */
      i++;
      k = 0;
      while (((i+k) < (MAX_FILENAME+MAX_HASH_SIZE+16)) && (k<MAX_HASH_SIZE) 
	     && (l->line[i+k]!='\0'))
	{
	  l->hash[k] = l->line[i+k];
	  k++;
	}
      l->hash[k]= (char) 0; /* fin de la chaine de caractre du hash */

      /* extraction de la chaine "MD5(" */
      k=0; 
      while ((k<MAX_HASH_SIZE) && (l->line[k]!='('))
	k++;

      /* extraction du reste de la chaine (jusqu' la fin)
	 k est sur la premire lettre du nom de fichier
	 i est positionn sur la parenthse fermante */
      i -= 3;
      k++;
      j = 0;
      while ((j<MAX_FILENAME) && ((k+j) < (MAX_FILENAME+MAX_HASH_SIZE+16)) 
	     && (k+j<i))
	{
	  l->nom[j] = l->line[k+j];
	  j++;
	}
      l->nom[j] = (char) 0;
   /*   while (j>0 && l->nom[j]!=')')
	{
	  j--;
	  l->nom[j] = (char) 0;
	  }
   */
    }
  else 
    {
      l->line[0] = (char) 0;
      l->nom[0] = (char) 0;
      l->hash[0] = (char) 0;
      l->ok = FALSE;
    }

 return l;
}

l_hash *load_list_hash_set_file(char *filename, l_hash *question, ligne *l)
{ 
  /* On charge le fichier md5 que l'on dsire vrifier dans une liste
     de toutes les faons il faudra vrifier tous les fichiers, alors
     Autant se simplifier la vie ! */
  int filenum=0;
  int nb=0;
  l_hash *tete= question;


  l->ok = TRUE;

  if ((filenum=open(filename, O_RDONLY))>0)
    {
      if (opt.debug == TRUE)
	fprintf(stdout, "%s opened\n", filename);
      while (l->ok == TRUE)
	{
	  l = lit_ligne(filenum, l);
	  if (l->ok == TRUE)
	    {
	      nb++;
	      tete = cons_l_hash(tete, filename, l->nom, l->hash);
	    }
	}
      close(filenum);
    }
  else  /* du open */
    {
      perror("load_list_hash_set_file(\"/open\")");
    }
	  
  if (opt.debug == TRUE)
    fprintf(stdout,"%s loaded (%d)\n",filename,nb);

  return tete;
}

st_rech *load_hash_set_file(char *filename,st_rech *sigs, ligne *l)
{ 
  /* on charge les fichiers contenant les signatures directement dans les 
     structures ici on utilise la structure de recherche rapide (on va 
     faire plusieurs dizaines de miliers de parcours pour rechercher des 
     chaines, autant que ce soit le plus rapide possible !). */

  int filenum=0;
  int nb=0;
  int real=0;
  st_rech *res = NULL;

  l->ok = TRUE;
 
  if ((filenum=open(filename, O_RDONLY))>0)
    {
      if (opt.debug == TRUE)
	fprintf(stdout, "%s opened\n", filename);
      while (l->ok == TRUE)
	{
	  l = lit_ligne(filenum, l);
	  if (l->ok == TRUE)
	    {
	      nb++;
	      /* if (opt.debug == TRUE)
		 fprintf(stdout, "?%s\t%s\t%s (%p)", filename, l->nom,
		         l->hash, sigs); */
	      res = cons_st_rech(sigs, filename, l->nom, l->hash,
				 MAX_INDIRECT, opt.unique);
	      if (res != NULL)
		{
		  real +=1;
		  sigs = res;
		  /*  if (opt.debug==TRUE)
		      fprintf(stdout," +\n"); */
		}
	      else
		{
		  /* if (opt.debug==TRUE)
		     fprintf(stdout," -\n"); */
		}
	    }
	}
      close(filenum);
    }
  else  /* du open */
    {
      perror("load_hash_set_file(\"/open\")");
    }	
 
  if (opt.debug == TRUE)
    fprintf(stdout,"%s loaded (%d) --> %d && %p\n", filename, nb, real, sigs);

  return sigs;
}


st_rech *load_hash_sets (st_rech *sigs)
{
  DIR *Rep = NULL;
  struct dirent *namelist = NULL;
  struct stat * stat_buffer = NULL;
  char * filename = NULL;
  int len_path=0;
  ligne *l=NULL;

  /* initialisations */

  stat_buffer = (struct stat*) malloc (sizeof(struct stat));
  if (stat_buffer != NULL)
    {
      len_path = strlen(opt.hash_sets_path);

      Rep = opendir(opt.hash_sets_path);

      if ( Rep==NULL )
	{
	  perror("opendir(\"/opt.hash_sets_path\")");
	  exit( FALSE );
	}

      namelist = readdir(Rep);

      l = (ligne*) malloc (sizeof(ligne));
      if (l != NULL)
	{
	  l->line = (char*) calloc ((MAX_FILENAME+MAX_HASH_SIZE+16), sizeof(char));
	  l->nom = (char*) calloc (MAX_FILENAME, sizeof(char));
	  l->hash = (char*) calloc (MAX_HASH_SIZE, sizeof(char));
	  l->ok = FALSE;

	  if (l->line !=NULL && l->nom != NULL && l->hash != NULL)
	    { /* on a pu allouer la mmoire suffisante pour continuer
		 La partie qui fait tout !!!                           */
	      while (namelist!=NULL ) 
		{
		  filename = (char*) malloc 
		    ((3+len_path+strlen(namelist->d_name))*sizeof(char));
		  sprintf(filename, "%s%s%c", opt.hash_sets_path, 
			  namelist->d_name, '\0');
		  stat(filename, stat_buffer); 
		  /* si c'est des fichiers rguliers (ie pas des rpertoires) */
		  if (S_ISREG(stat_buffer->st_mode)) 
		    {
		      sigs = load_hash_set_file(filename,sigs,l);
		    }
		  namelist = readdir(Rep);
		  free(filename);
		}
	      /* fin de la partie qui fait tout ! */
              /* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
	      /* !! on ne libre pas la mmoire !! */
	    }
	} 
      else
	{
	  if (opt.debug == TRUE)
	    fprintf(stderr, "load_hash_set_file - Erreur d'allocation\
 mmoire : l : %p\n", l);
	}
      

      free(stat_buffer);
    }
  else
    {
      if (opt.debug == TRUE)
	fprintf(stderr,"load_hash_sets : Erreur d'Allocation mmoire\
 : stat_buffer\n"); 
    }

  return sigs;
}


void affiche_liste(l_hash *liste)
{
  l_hash *l=liste;

  while (l != NULL)
    {
      fprintf(stdout, "%s\t%s\t%s\n", l->nom_hash_set, l->nom_fichier, l->hash);
      l = l->suivant;
    }
}

void affiche_longueur(st_rech *sigs)
{
  int i=0;

  for(i=0; i<16; i++)
    {
      if (sigs->ind[i]!=NULL)
	affiche_longueur(sigs->ind[i]);
      else
	{
	  fprintf(stdout,"Liste %d a une longueur de %d\n", i, 
		  longueur_l_hash(sigs->tab[i]));
	  
	}
    }
}


void affiche_stats(stat_st_rech *s)
{
  if ((s->nbdelistes-s->nbdelistesvides) != 0)
    {
      fprintf(stdout,"Nombre d'tages : %d\nNombre d'indirections : %d\n\
Nombre de fichiers chargs : %d\nNombre de listes cres : %d\n\
Nombre de liste vides : %d\nLongueur maximale d'une liste : %d\n\
Longueur minimale d'une liste : %d\nLongueur moyenne d'une liste : %d\n",
	  MAX_INDIRECT, s->nbindirections,
	  s->nbdefichiers, s->nbdelistes,
	  s->nbdelistesvides, s->longueurmax, 
	  s->longueurmin, s->nbdefichiers/(s->nbdelistes-s->nbdelistesvides));
    }
  else
    {
      fprintf(stdout,"Nombre d'tages : %d\nNombre d'indirections : %d\n\
Nombre de fichiers chargs : %d\nNombre de listes cres : %d\n\
Nombre de liste vides : %d\nLongueur maximale d'une liste : %d\n\
Longueur minimale d'une liste : %d\n",
	  MAX_INDIRECT, s->nbindirections,
	  s->nbdefichiers, s->nbdelistes,
	  s->nbdelistesvides, s->longueurmax, 
	  s->longueurmin);
    }
}


void free_list(l_hash *liste) /* a descend et a libre en remontant ! */
{
  if (liste!=NULL)
    {
      free_list(liste->suivant);
      free(liste);
    }
}


void free_hash_set(st_rech *sigs)
{
 int i=0;
 
 if (sigs != NULL)
   {
     for(i=0; i<16; i++)
       {
	 if (sigs->ind[i]!=NULL)
	   free_hash_set(sigs->ind[i]);
	 else	
	   free_list(sigs->tab[i]);
       }
   }
}

void affiche_element_liste(l_hash *liste,l_hash *liste2,int c1,int c2, int c3)
{
  /* affiche l'lment de la liste en focntion des options d'affichage */
  char fin='\n';
 

  if (opt.print == TRUE)
    fin = '\0';
  else
    fin = '\n';

  if (c1+c2+c3 > 1)
    {  /* on en affiche au moins deux ! */
      if (c1 == 1)
	fprintf(stdout, "%s\t", liste->nom_hash_set);
      if (c2 == 1)
	fprintf(stdout, "%s", liste2->nom_fichier); 
      /* parce que il peut tre a la fois premier et dernier ! */
      if (c3 == 1)
	fprintf(stdout,"\t%s",liste->hash);
    }
  else
    { /* on en affiche qu'un seul ! */
      if (c1 == 1)
	fprintf(stdout, "%s", liste->nom_hash_set);
      if (c2 == 1)
	fprintf(stdout, "%s", liste2->nom_fichier);
      if (c3 == 1)
	fprintf(stdout, "%s", liste->hash);
    }
  if (opt.aff != 0)
    fprintf(stdout, "%c", fin);
  
}



void match_it(l_hash *question, st_rech *sigs)
{
  // fonction a revoir !!!
 
  l_hash *trouve = NULL;
  l_hash *tete = question;
  int matched = 0;
  int rejected = 0;

  char nb = (char) 0;
  int c1 = 0;
  int c2 = 0;
  int c3 = 0;

  nb = opt.aff;
  
  if ((nb-4) >= 0)
   {
     c1 = 1;
     nb = nb-4;
   }
 
  if ((nb-2) >= 0)
    {
      c2 = 1;
      nb = nb-2;
    }
  
  if ((nb-1) >= 0)
    c3 = 1;

  if (opt.debug==TRUE)
    fprintf(stdout, "Display columns : %d %d %d\n", c1, c2, c3);
  while (question!=NULL)
    {
      trouve = rech_st_rech(sigs, question, MAX_INDIRECT);

      if (trouve!=NULL)
	{
	  if (opt.match == TRUE)
	    affiche_element_liste(trouve, question, c1, c2, c3);
	  matched++;
	}
      else
	{
	  if (opt.reject == TRUE)
	    affiche_element_liste(question, question, c1, c2, c3);
	  rejected++;
	} 

      if (opt.debug==TRUE)
	fprintf(stdout,"Matched : %d\tRejected : %d\n", matched, rejected);
      question = question->suivant;
    }
  /* if (opt.stat == TRUE) */
    fprintf(stderr, "Trouvs : %d\tInconnus : %d\tTotal : %d\n", matched, 
	    rejected, longueur_l_hash(tete));

  
}



int main (int argc, char ** argv)
{

  st_rech *sigs = NULL;
  l_hash *question = NULL;
  int c = 0;
  int help = 0;
  int val = 0;
  int nb = 0;
  char *filename = NULL;
  struct stat *stat_buffer = NULL;
  stat_st_rech *statistiques = NULL;
  ligne *l = NULL;

  /* Initialisations des paramtres                                    */
  opt.debug = FALSE;       /* Par dfaut on debug pas !                */
  opt.match = TRUE;        /* Par dfaut on affiche ceux qui sont bons */
  opt.reject = FALSE;      /* et pas les autres !                      */
  opt.stat = FALSE;        /* On affiche pas les stats                 */
  opt.print = FALSE;       /* On affiche avec des sauts de ligne       */
  opt.unique = FALSE;
  opt.aff = (char) 7;

  
  val = TRUE;   /* Valeur de retour du programme                       */
  help = FALSE; /* Aide, non demande par dfaut (--help)              */
  c = 0;        /* Nombre de paramtres de la ligne de commande        */
 
  opt.hash_sets_path = (char *) malloc 
                                ((strlen(HASH_SETS_PATH)+1)*sizeof(char));
  if (opt.hash_sets_path != NULL)
    {
      sprintf(opt.hash_sets_path, "%s%c", HASH_SETS_PATH, '\0');
  
      statistiques = (stat_st_rech*) malloc (sizeof(stat_st_rech)); 
  
      if (statistiques != NULL && sizeof(*statistiques) == sizeof(stat_st_rech))
	{

	  statistiques->nbdefichiers = 0;
	  statistiques->nbdelistes = 0;
	  statistiques->nbindirections = 0;
	  statistiques->nbdelistesvides = 0;
	  statistiques->longueurmin = MAX_INT;
	  statistiques->longueurmax = 0;

	  prog_name = (char*) malloc (sizeof(char)*256);
	  if (prog_name != NULL)
	    {
	      prog_name = argv[0];


	      while ((c = getopt_long (argc, argv, "vhdmrtc0", long_options, NULL)) != -1)
		{
		  switch (c)
		    {
		    case 0:			/* Long option. */
		      break;
	  
		    case 'v':
		      version();
		      break;
	 
		    case 'h':
		      help = TRUE; 
		      break;
	  
		    case 's':
		      if (optarg)
			{
			  free(opt.hash_sets_path);
			  opt.hash_sets_path = (char *) malloc 
			    ((strlen(optarg)+1)*sizeof(char));
			  if (opt.hash_sets_path != NULL)
			    {
			      if (sscanf(optarg, "%s", opt.hash_sets_path) < 1)
				{
				  free(opt.hash_sets_path);
				  opt.hash_sets_path = (char *) malloc 
				    ((strlen(HASH_SETS_PATH)+1)*sizeof(char));
				  if (opt.hash_sets_path != NULL)
				    sprintf(opt.hash_sets_path, "%s%c", 
					    HASH_SETS_PATH, '\0');
				  else
				    {
				      if (opt.debug == TRUE)
					fprintf(stderr, "main - Erreur\
 d'Allocation mmoire : opt.hash_sets_path\n"); 
				    }
				}
			    }
			  else
			    {
			      if (opt.debug == TRUE)
				fprintf(stderr,"main - Erreur\
 d'Allocation mmoire : opt.hash_sets_path\n"); 
			    }
			}
		      break; 
		    case 'c':
		      if (optarg)
			{
			  if (sscanf(optarg, "%d", &nb) < 1)
			    {
			      opt.aff = (char) 7;
			    }
			  opt.aff = (char) nb;
			}
		      else
			opt.aff = (char) 7;
		      break; 

		    case 'd':
		      opt.debug = TRUE; 
		      break;

		    case 'm':
		      opt.match = TRUE; 
		      opt.reject = FALSE;
		      break;

		    case 'r':
		      opt.reject = TRUE; 
		      opt.match = FALSE;
		      break;

		    case 't':
		      opt.stat = TRUE; 
		      break;

		    case '0':
		      opt.print = TRUE; 
		      break;

		    case 'u':
		      opt.unique = TRUE; 
		      break;

		    default:
		      usage(1);
		    }
		}
  
	      if (opt.debug == TRUE)
		fprintf(stdout, "Taille de sigs : %d\n", sizeof(*sigs));

	      /* les arguments qui ne sont pas dans ceux dfinis : */
	      if (optind < argc)
		{
		  /* initialisation du hash set */
		  sigs = load_hash_sets(sigs);

		  if (opt.debug == TRUE)
		    fprintf(stdout,"All list has been loaded\n");

		  while (optind < argc)
		    {
		      filename = (char*) malloc 
			(sizeof(char)*strlen(argv[optind])+2);
		      stat_buffer = (struct stat*) malloc (sizeof(struct stat));
		     if (opt.debug == TRUE)
		       fprintf(stdout,"while optind(%d) < argc (%d)\n",optind,argc);
		      if (filename != NULL && stat_buffer != NULL)
			{
			  sprintf(filename, "%s%c",argv[optind], '\0');
			  optind++;
			  stat(filename,stat_buffer); 
			  if (S_ISREG(stat_buffer->st_mode)) 
			    { 
			      /* si c'est un fichier rgulier (ie pas des 
				rpertoires)
				Pas mal le NULL planqu ici !             */
			      question = NULL;

			      if (opt.debug == TRUE)
				fprintf(stdout, "Comparison list is being loaded\n");

			      l = (ligne*) malloc (sizeof(ligne));
			      if (l != NULL)
				{
				  l->line = (char*) calloc 
				    ((MAX_FILENAME+MAX_HASH_SIZE+16), sizeof(char));
				  
				  l->nom = (char*) calloc 
				    (MAX_FILENAME,sizeof(char));
				  
				  l->hash = (char*) calloc 
				    (MAX_HASH_SIZE,sizeof(char));
				  
				  l->ok = FALSE;
				  if (l->line !=NULL && l->nom != NULL 
				      && l->hash != NULL)
				    {
				      
				     question = 
				       load_list_hash_set_file(filename, 
							       question, l);
				      
				     /* !! On ne libre pas la mmoire !!
					free(l->line);
				        free(l->nom);
				        free(l->hash);
				     */
				    }
				  /* free(l); */
				}

			      /* A partir de l on peut faire le traitement ! */
			      if (opt.debug == TRUE)
				{
				  fprintf(stdout, 
					  "Comparison list loaded : %d files\n",
					  longueur_l_hash(question));
				  fprintf(stdout, "Begining match function\n");
				}

			      match_it(question,sigs);

			      if (opt.debug == TRUE)
				fprintf(stdout, "Match function done\n");
			      
			    }
			  /* free(filename);
			     free(stat_buffer); */
			}
		      else
			{
			  if (opt.debug == TRUE)
			    fprintf(stderr, "main - Erreur d'Allocation mmoire\
 : filename %p\tstat_buffer :%p\n", filename, stat_buffer); 
			}
		    }

		  if (opt.stat == TRUE)
		    {
		      if (opt.debug == TRUE)
			fprintf(stdout, 
				"Begining compute_stat_st_rech function\n");
		      compute_stat_st_rech(statistiques, sigs);
		      if (opt.debug == TRUE)
			fprintf(stdout, 
				"compute_stat_st_rech function Done\n");
		      affiche_stats(statistiques);      
		    }
		}
	      else
		{
		  if (help == TRUE)
		    usage(0);
		  else
		    usage(1);
		}
    
	      /* free(prog_name); */
	    }
	}
      else
	{
	  if (opt.debug == TRUE)
	    fprintf(stderr, "main - Erreur d'Allocation mmoire : sigs\
 %p\tstatistiques :%p\n", sigs, statistiques); 
	}
    }
  else
    {
      if (opt.debug == TRUE)
	fprintf(stderr, "main - Erreur d'Allocation mmoire :\
 opt.hash_sets_path\n"); 
    }
  return val;
}
