/******************************************************************************
*                                                                             *
* Librairie de fonctions pour la gestion des structures l_hash et st_rech     *
* Auteur : Delhomme Olivier, 26 06 2003                                       *
*                                                                             *
* Version 0.0.2 (31.05.2003)                                                  *
*    - Normalisation C                                                        *
******************************************************************************/

#include <stdlib.h>
#include <string.h>
#include "delh.h"

l_hash *cons_l_hash(l_hash *liste, char *nom_hash_set, char *nom_fichier, 
                    char *hash)
{
  /* construction en tte d'une liste. */

  l_hash *l = NULL;

  l = (l_hash*) calloc (1,sizeof(l_hash));
  
  if (l && sizeof(*l) == sizeof(l_hash))
    {
      l->nom_hash_set = (char*) calloc (1+strlen(nom_hash_set), sizeof(char));
      if (l->nom_hash_set != NULL && nom_hash_set != NULL)
	memcpy(l->nom_hash_set, nom_hash_set, 1+strlen(nom_hash_set));

      l->nom_fichier = (char*) calloc (1+strlen(nom_fichier), sizeof(char));
      if (l->nom_fichier != NULL && nom_fichier != NULL)
	memcpy(l->nom_fichier, nom_fichier, 1+strlen(nom_fichier));

      l->hash = (char*) calloc (1+strlen(hash), sizeof(char));
      if (l->hash != NULL && hash != NULL)
	memcpy(l->hash, hash, 1+strlen(hash));
  
      l->suivant = liste;
      return l;
    }
  else
    return (NULL);
}

l_hash *free_l_hash(l_hash *liste)
{
  /* libration en tte */

  l_hash *l = NULL;

  if (liste != NULL)
    {
      l = liste->suivant;

      if (liste->nom_hash_set != NULL)
	free(liste->nom_hash_set);

      if (liste->nom_fichier != NULL)
	free(liste->nom_fichier);

      if (liste->hash != NULL)
	free(liste->hash);
      
      free(liste);
    }

  return l;
}

l_hash *rech_l_hash(char *hash, l_hash *connue)
{
  /* recherche dans la liste connue (recherche linaire O(n)) */
  /* la comparaison s'effectue avec le 'hash' */
 
  if (hash != NULL)
    {
      int ok = FALSE;
      l_hash *l = connue;
      
      while (l != NULL && ok == FALSE)
	{
	  /*	  fprintf(stdout,"rech_l_hash, l->hash : %p et %p\n",l,l->suivant); */
	  if (strcmp(hash, l->hash) == 0)
	    ok = TRUE;
	  else
	    l = l->suivant;
	}
      return l;
    }
  else 
    return NULL;
}

int longueur_l_hash(l_hash *liste)
{
  if (liste !=  NULL)
    return (1+longueur_l_hash(liste->suivant));
  else
    return (0);
}

/*----------------------------------------------------------------------------*/

st_rech *NULL_st_rech()
{
  /* Obtention d'une nouvelle structure. */
  st_rech *new_sig = NULL;
  int i = 0;

  new_sig = (st_rech*) calloc (1, sizeof(st_rech));

  if (new_sig != NULL)
    {
      for (i=0; i<16; i++)
	{
	  new_sig->ind[i] = NULL;
	  new_sig->tab[i] = NULL;
	}
    }

  return  new_sig;
}

int clef_hachage_st_rech(char c)
{
  /* calcul de la cl de hachage :                 */
  /* on utilise les proporits de l'hexadcimal   */
  /* insensible  la case                          */
  switch (c)
    {
    case '1':
      return 1;
      break;

    case '2':
      return 2;
      break;

    case '3': 
      return 3;
      break;

    case '4': 
      return 4;
      break;

    case '5': 
      return 5;
      break;

    case '6': 
      return 6;
      break;

    case '7': 
      return 7;
      break;

    case '8': 
      return 8;
      break;

    case '9': 
      return 9;
      break;

    case 'a':
    case 'A':
      return 10;
      break;

    case 'b':
    case 'B':
      return 11;
      break;

    case 'c':
    case 'C':
      return 12;
      break;

    case 'd':
    case 'D': 
      return 13;
      break;

    case 'e':
    case 'E':
      return 14;
      break;

    case 'f': 
    case 'F':
      return 15;
      break;

    default: /* le cas 0 et tous les autres cas ! */
      return 0;
    }

  return 0; /* si ca ne va pas on se rattrape sur 0 !!*/
}


int trouv_st_rech(st_rech *sig, char *nom_hash_set, char *nom_fichier, 
                   char *hash, int niveau, int ou, int collision)
{
  /* insertion dans une liste, avec gestion des collisions 
     si collision == TRUE                                       */

  int clef = 0 ;

  if (ou < niveau) /* on rcurse */
    {
      if (hash != NULL)
	{
	  clef = clef_hachage_st_rech(hash[ou]);
	  
	  if (sig->ind[clef] == NULL)
	    sig->ind[clef] = NULL_st_rech();
	  
	  if (sig->ind[clef] != NULL)
	    return trouv_st_rech(sig->ind[clef], nom_hash_set, nom_fichier, 
                                 hash, niveau, ou++, collision);
	  else
	    return FALSE;
	}
      else return FALSE;
    }
  else /* on insre (avec gestion des collisions par une liste) */
    {
      if (hash != NULL)
	{
	  clef = clef_hachage_st_rech(hash[ou]);
	  if (collision == TRUE)
	    {
	      if (rech_l_hash(hash, sig->tab[clef]) == NULL)
		{
		  sig->tab[clef] = cons_l_hash(sig->tab[clef], nom_hash_set,
					       nom_fichier, hash);
		  return TRUE;
		}
	      else
		return FALSE;
	    }
	  else
	    {
	      sig->tab[clef] = cons_l_hash(sig->tab[clef], nom_hash_set, 
					   nom_fichier, hash);
	      return TRUE;
	    }
	}
      else
	return FALSE;
    }
}


st_rech *cons_st_rech(st_rech *sig, char *nom_hash_set, char *nom_fichier, 
                   char *hash, int niveau, int collision)
{
  /* on cherche  construire la structure 'sig' pour 
     y placer la l'lment de liste 'liste' (non vide si possible !!)
     niveau == valeur maximale d'indirections
     !! retourne NULL si l'lment n'a pas t insr !!               */

  int res = TRUE;  /* optimiste :) */

  if (sig == NULL)
    sig = NULL_st_rech();
  
  if (sig != NULL)
    res = trouv_st_rech(sig, nom_hash_set, nom_fichier, hash, niveau, 0, collision);
  
  if (res == TRUE)
    return sig;
  else
    return NULL;
}


l_hash *look_for_hash(st_rech *sig, l_hash *liste, int niveau, int ou)
{
  /* recherche la place o doit se trouver la clef
     recherche par arbre en O(log16(n)) puisque la clef permet 16 positions
     sig :  la structure de hash
     liste : contient le hash  rechercher, 
     ou : un nombre gal  0 au dpart, et qui augmente
     niveau : le nombre maxi d'indirections
  */
 
  /* j'utilise cette variable pour pouvoir tester les structures */
  int clef = 0;  

  if (ou < niveau)
    {
      if (liste->hash != NULL)
	{
	  clef = clef_hachage_st_rech(liste->hash[ou]);
	  if (sig->ind[clef] != NULL)
	    return look_for_hash(sig->ind[clef], liste, niveau, ou++);
	  else
	    return NULL;
	}
      else
	return NULL;
    }
  else
    {  /* on est au bout (ou == niveau) */
      if (liste->hash != NULL)
	{
	  clef = clef_hachage_st_rech(liste->hash[ou]);
	  if (sig->tab[clef] != NULL)
	    return rech_l_hash(liste->hash,sig->tab[clef]);
	  else
	    return NULL;
	}
      else
	return NULL;
    }
  return NULL;
}

l_hash *rech_st_rech(st_rech *sig, l_hash *liste, int niveau)
{ 
  /* recherche le hash contenu dans l'lment de liste 'liste' dans
     l'ensemble de hashs 'sigs' et retourne l'lment de liste
     correspondant. La recherche se base sur le hash. */
  
  if ((liste != NULL) && (sig != NULL))
    return look_for_hash(sig, liste, niveau, 0);
  else
    return NULL;
}


void compute_stat_st_rech(stat_st_rech *s, st_rech *sig)
{
  int i=0;
  int l=0;
  
  if (sig != NULL && s !=NULL)
    {
      for (i=0; i<16; i++)
	{
	  if (sig->tab[i] != NULL)
	    {
	      l = longueur_l_hash(sig->tab[i]);
	      s->nbdefichiers += l;
	      s->nbdelistes += 1;
	      if (l != 0)
		{
		  if (s->longueurmax < l)
		    s->longueurmax = l;
		  if (s->longueurmin > l)
		    s->longueurmin = l;
		}
	      else
		s->nbdelistesvides += 1;
	    }
	  if (sig->ind[i] != NULL)
	    {
	      s->nbindirections += 1;
	      compute_stat_st_rech(s,sig->ind[i]);
	    }
	}
    }
}
